/*******************************************************************************
Copyright © 2024 Sebastian Roith

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*******************************************************************************/

(function() {

/***[ SqlChecker.js ] *************************/
/** Basic class for checking possibilities of sql queries and results tables */
class SqlChecker {
    constructor(
        objAtts,        // Object containing values of object to create
        arrImportKeys   // Optional: Array of keys to import from `objAtts`
    ) {
        if (Number.isFinite(objAtts.fine) && objAtts.fine >= 0)
            this.fine = objAtts.fine;
        if (objAtts.msg)
            this.msg = objAtts.msg;
        if (objAtts.unlock)
            this.unlock = true;
        if (objAtts.inverted)
            this.inverted = true;
        if (objAtts.accept !== undefined && (objAtts.accept ? true : false) != this.accept)
            this.accept = objAtts.accept ? true : false;
        if (arrImportKeys instanceof Array)
            arrImportKeys.forEach(key => {
                if (objAtts[key] !== undefined)
                    this[key] = objAtts[key];
            });
    }

    /**
    * Checks whether data triggers this rule
    * @return boolean       Does data trigger this rule?
    */
    exec(
        data    // Object like { scope: "SqlScope instance", result: "SqlTable instance", code: "String", tokens: "Array of SqlBasisToken instances", query: "SqlExpression instance" }
    ) {
        return true;
    }

    /**
    * Checks whether data triggers this rule with respect to `this.inverted`
    * @return boolean       Does data trigger this rule?
    */
    execute(
        data    // Object like { scope: "SqlScope instance", result: "SqlTable instance", code: "String", tokens: "Array of SqlBasisToken instances", query: "SqlExpression instance" }
    ) {
        return this.exec(data) != this.inverted;
    }
}

/** Fine points gained if exec returns true */
SqlChecker.prototype.fine = 0;

/** Message to display if exec returns true */
SqlChecker.prototype.msg = '';

/** Is level unlocked if exec returns true? */
SqlChecker.prototype.unlock = false;

/** Does passing test mean acceptance? */
SqlChecker.prototype.accept = true;

/** Is check negated? */
SqlChecker.prototype.inverted = false;

/** Types of SqlChecker subclasses corresponding to type values */
SqlChecker.types = {
    "accept": SqlChecker
};

/**
* Create SqlChecker instance described by object `data`
* @return SqlChecker    Created instance (or null)
*/
SqlChecker.fromObject = function(
    data    // Object describing wanted instance
) {
    if (data instanceof Object && SqlChecker.types[data.type])
        return new SqlChecker.types[data.type](data);
    return null;
}

/***[ SqlCheckerAttributeCount.js ] *************************/

class SqlCheckerAttributeCount extends SqlChecker {
    constructor(
        objAtts     // Object containing values of object to create
    ) {
        super(objAtts, ['value', 'min', 'max']);
    }

    /**
    * Checks whether data triggers this rule?
    * @return boolean       Does data trigger this rule?
    */
    exec(
        data    // Object like { scope: "SqlScope instance", result: "SqlTable instance", code: "String", tokens: "Array of SqlBasisToken instances", query: "SqlExpression instance" }
    ) {
        if (!(data.result instanceof SqlTable))
            return true;
        const atts = data.result.atts.length;
        if (this.value >= 0 && this.value != atts)
            return true;
        if (atts < this.min || atts > this.max)
            return true;
        return false;
    }
}

// If positive: Exact number of required attributes
SqlCheckerAttributeCount.prototype.value = -1;

// Minimum number of required attributes
SqlCheckerAttributeCount.prototype.min = 0;

// Maximum accepted number of attributes
SqlCheckerAttributeCount.prototype.max = Infinity;


SqlCheckerAttributeCount.prototype.accept = false;
SqlCheckerAttributeCount.prototype.msg = 'Anzahl an Attributen in der Ergebnistabelle ist falsch';
SqlCheckerAttributeCount.prototype.fine = 1;

SqlChecker.types['attributeCount'] = SqlCheckerAttributeCount;

/***[ SqlCheckerAttributeName.js ] *************************/
/** Ensure predefined attribute name */
class SqlCheckerAttributeName extends SqlChecker {
    constructor(
        objAtts     // Object containing values of object to create
    ) {
        super(objAtts, ['value', 'index']);
    }

    /**
    * Checks whether data triggers this rule?
    * @return boolean       Does data trigger this rule?
    */
    exec(
        data    // Object like { scope: "SqlScope instance", result: "SqlTable instance", code: "String", tokens: "Array of SqlBasisToken instances", query: "SqlExpression instance" }
    ) {
        if (!(data.result instanceof SqlTable))
            return true;
        const atts = data.result.atts;
        if (this.index < 0)
            return atts.reduce((prev, { name }) => prev || name == this.value, false);
        if (!(atts[this.index] instanceof Object))
            return true;
        if (atts[this.index].name != this.value)
            return true;
        return false;
    }
}

// Name of attribute in question
SqlCheckerAttributeName.prototype.value = '';

// Index of attribute in question (if negative: index does not matter)
SqlCheckerAttributeName.prototype.index = -1;


SqlCheckerAttributeName.prototype.accept = false;
SqlCheckerAttributeName.prototype.msg = 'Falscher Attributname in Ergebnistabelle';
SqlCheckerAttributeName.prototype.fine = 1;

SqlChecker.types['attributeName'] = SqlCheckerAttributeName;

/***[ SqlCheckerCompareQuery.js ] *************************/

class SqlCheckerCompareQuery extends SqlChecker {
    constructor(
        objAtts     // Object containing values of object to create
    ) {
        super(objAtts, ['value', 'substring', 'casesensitive', 'noNormalize']);
    }

    /**
    * Checks whether data triggers this rule?
    * @return boolean       Does data trigger this rule?
    */
    exec(
        data    // Object like { scope: "SqlScope instance", result: "SqlTable instance", code: "String", tokens: "Array of SqlBasisToken instances", query: "SqlExpression instance" }
    ) {
        if (!(data.query instanceof SqlExpression))
            return true;
        let query = this.noNormalize ? data.code : data.query.toString();
        let target = this.value;
        if (!this.casesensitive) {
            query = query.toUpperCase();
            target = target.toUpperCase();
        }
        if (this.substring)
            return (query.indexOf(target) == -1);
        return (query != target);
    }
}

// Text to compare query with
SqlCheckerCompareQuery.prototype.value = "";

// May `value` just be a substring? Otherwise it has to match query completely
SqlCheckerCompareQuery.prototype.substring = false;

// Compare case sensitive?
SqlCheckerCompareQuery.prototype.casesensitive = false;

// Use query inputed by user directly instead of normalized query?
SqlCheckerCompareQuery.prototype.noNormalize = false;

SqlCheckerCompareQuery.prototype.accept = false;
SqlCheckerCompareQuery.prototype.msg = 'Die eingegebene Abfrage entspricht nicht den Vorgaben';
SqlCheckerCompareQuery.prototype.fine = 1;

SqlChecker.types['compareQuery'] = SqlCheckerCompareQuery;

/***[ SqlCheckerKeyword.js ] *************************/

class SqlCheckerKeyword extends SqlChecker {
    constructor(
        objAtts     // Object containing values of object to create
    ) {
        super(objAtts, ['value', 'min', 'max']);
    }

    /**
    * Checks whether data triggers this rule?
    * @return boolean       Does data trigger this rule?
    */
    exec(
        data    // Object like { scope: "SqlScope instance", result: "SqlTable instance", code: "String", tokens: "Array of SqlBasisToken instances", query: "SqlExpression instance" }
    ) {
        if (!(data.tokens instanceof Array))
            return true;
        const keyword = this.value.toUpperCase();
        const count = data.tokens.reduce((prev, token) => prev + (token.type == SqlBasisToken.TOKEN_KEYWORD && token.content.toUpperCase() == keyword), 0);
        if (count < this.min || count > this.max)
            return true;
        return false;
    }
}

// Text content of keyword
SqlCheckerKeyword.prototype.value = "";

// Minimum number of required occurences
SqlCheckerKeyword.prototype.min = 0;

// Maximum accepted number of occurences
SqlCheckerKeyword.prototype.max = Infinity;


SqlCheckerKeyword.prototype.accept = false;
SqlCheckerKeyword.prototype.msg = 'Ein Schlüsselwort kommt nicht in der geforderten Anzahl vor';
SqlCheckerKeyword.prototype.fine = 1;

SqlChecker.types['keyword'] = SqlCheckerKeyword;

/***[ SqlCheckerSameResult.js ] *************************/
/** Ensure predefined results table */
class SqlCheckerSameResult extends SqlChecker {
    constructor(
        objAtts     // Object containing values of object to create
    ) {
        super(objAtts, ['value', 'checkOrder']);
    }

    /**
    * Checks whether data triggers this rule?
    * @return boolean       Does data trigger this rule?
    */
    exec(
        data    // Object like { scope: "SqlScope instance", result: "SqlTable instance", code: "String", tokens: "Array of SqlBasisToken instances", query: "SqlExpression instance" }
    ) {
        if (!(data instanceof Object) || !(data.scope instanceof SqlScope))
            return false;
        if (!(data.result instanceof SqlTable))
            return true;
        const tokens = SqlBasisToken
            .tokenize(this.value)
            .filter(token => [SqlBasisToken.TOKEN_WHITESPACE, SqlBasisToken.TOKEN_COMMENTLINE, SqlBasisToken.TOKEN_COMMENTMULTI].indexOf(token.type) == -1);
        const parts = [];
        try {
            if (!SqlExpressionSelect.fromTokens(tokens, parts) || !parts.length) {
                console.warn("Invalid SQL query to compare with:", data.code);
                return false;
            }
            const query = parts[0];
            const scope = data.scope;
            query.calcValue(scope);
            let target = scope.tableResult;
            let source = data.result;
            if (target.atts.length != source.atts.length || target.data.length != source.data.length)
                return true;
            if (target.atts.reduce((prev, { name }, index) => prev || name != source.atts[index].name, false))
                return true;
            target = target.toJSON().map(row => JSON.stringify(row));
            source = source.toJSON().map(row => JSON.stringify(row));
            target.shift();
            source.shift();
            if (this.checkOrder) {
                return target.reduce((prev, row, index) => prev || row != source[index], false);
            } else {
                return target.reduce((prev, row) => {
                    if (prev)
                        return prev;
                    const index = source.indexOf(row);
                    if (index < 0)
                        return true;
                    source.splice(index, 1);
                    return false;
                }, false);
            }
        } catch(err) {
            console.warn(err);
        }
        return false;
    }
}

// SQL query having requested results table
SqlCheckerSameResult.prototype.value = '';

// Also check order of rows?
SqlCheckerSameResult.prototype.checkOrder = false;


SqlCheckerSameResult.prototype.accept = false;
SqlCheckerSameResult.prototype.msg = 'Falsche Ergebnistabelle';
SqlCheckerSameResult.prototype.fine = 1;

SqlChecker.types['sameResult'] = SqlCheckerSameResult;

/***[ SqlCheckerAnd.js ] *************************/

class SqlCheckerAnd extends SqlChecker {
    constructor(
        objAtts     // Object containing values of object to create
    ) {
        super(objAtts, ['value']);
        if (!(this.value instanceof Array) || !this.value.length)
            return;
        const firstChecker = SqlChecker.fromObject(this.value[0]);
        ['accept', 'msg', 'fine', 'unlock'].forEach(key => {
            if (this[key] === undefined)
                this[key] = firstChecker[key];
        });
    }

    /**
    * Checks whether data triggers this rule?
    * @return boolean       Does data trigger this rule?
    */
    exec(
        data    // Object like { scope: "SqlScope instance", result: "SqlTable instance", code: "String", tokens: "Array of SqlBasisToken instances", query: "SqlExpression instance" }
    ) {
        if (!(this.value instanceof Array) || !this.value.length)
            return false;
        return this.value.reduce((prev, json) => prev && SqlChecker.fromObject(json).execute(data), true);
    }
}

// JSON of SqlChecker instances to run
SqlCheckerAnd.prototype.value = [];

SqlCheckerAnd.prototype.accept = undefined;
SqlCheckerAnd.prototype.msg = '';
SqlCheckerAnd.prototype.fine = undefined;
SqlCheckerAnd.prototype.unlock = undefined;

SqlChecker.types['and'] = SqlCheckerAnd;

window.SqlChecker = SqlChecker;
})();
